import contextlib
import logging
import sys
import time

from lettuce import step

from selenium.common.exceptions import WebDriverException


from fattoush.driver.driver import Driver
from fattoush.util import filename_in_created_dir
from fattoush.namespace import per

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)
HANDLER = logging.StreamHandler()
HANDLER.setLevel(logging.INFO)
LOGGER.addHandler(HANDLER)


class Decorated(object):
    def __init__(self, fn, decorators):
        self.fn = fn
        self.func_code = fn.func_code
        self.decorators = decorators

    def __call__(self, *args, **kwargs):
        head, tail = self.decorators[0], self.decorators[1:]
        fn, args, kwargs = head(self.fn, *args, **kwargs)

        if tail:
            return self.__class__(fn, tail)(*args, **kwargs)

        return fn(*args, **kwargs)


class Step(object):
    def __init__(self, *decorated):
        self._decorated = decorated

    def decorator(self, fn, step_regex):
        decorated = Decorated(fn, self._decorated)

        return step(step_regex)(decorated)

    def __call__(self, step_regex):

        def decorator(fn):
            return self.decorator(fn, step_regex)

        return decorator

    def then(self, decorated):
        chain = self._decorated + (decorated, )

        return self.__class__(*chain)


@contextlib.contextmanager
def _screenshot_after(step):
    """
    Ensure that a screenshot is taken after the decorated step definition
    is run.
    """
    file_path = filename_in_created_dir(dir_name='logs', step=step, ext='png')

    try:
        yield
    except:
        exc_type, exc_value, exc_tb = sys.exc_info()

        time.sleep(1)

        browser = per.scenario.get('browser')

        if browser is not None:
            try:
                taken = browser.get_screenshot_as_file(file_path)
            except WebDriverException:
                taken = False

            if taken:
                LOGGER.info(
                    "captured screen shot to {}".format(file_path)
                )
            else:
                LOGGER.exception(
                    "could not capture screen shot to {}".format(file_path)
                )

        raise exc_type, exc_value, exc_tb
    else:
        browser = per.scenario.get('browser')
        if browser is None:
            return

        try:
            if browser.is_sauce:
                browser.get_screenshot_as_png()
            else:
                browser.get_screenshot_as_file(file_path)
        except WebDriverException:
            pass


@Step
def without_step(definition, _, *args, **kwargs):
    return definition, args, kwargs


@Step
def just_hashes(definition, step_obj, *args, **kwargs):
    return definition, (step_obj.hashes,) + args, kwargs


@Step
def screenshot(definition, step_obj, *args, **kwargs):
    with _screenshot_after(step_obj):
        return definition, (step_obj,) + args, kwargs


@screenshot.then
def with_wd(definition, step_obj, *args, **kwargs):
    wd = Driver.instance(step_obj)

    return definition, (step_obj, wd) + args, kwargs


@with_wd.then
def just_wd(definition, _, *args, **kwargs):
    return definition, args, kwargs



# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

"""
Hooks that are added to lettuce to enable fattoush to work in the
manner described in the module's documentation. Just import this
module (`fattoush.hooks`) to apply them.
"""

from . import feature, scenario, step



# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

from .driver import Driver
from .sauce import Sauce, Local, SauceInterface



# -*- coding: utf-8 -*-
# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

from fattoush import (
    util,
    step_just_wd,
)


def _send_to_search_box(wd, keys):
    search_box = wd.find_element_by_css_selector('form input[title=Search]')
    search_box.send_keys(keys)


@step_just_wd(u'I open "([^"]*)" in my web browser')
def i_open_url_in_my_web_browser(wd, url):
    wd.get(url)


@step_just_wd(u'I expect the url to start with "([^"]*)"')
def i_expect_the_url_to_start_with(wd, url):
    current_url = wd.current_url

    assert current_url.startswith(url), '{!r} did not start with {!r}'.format(
        current_url, url
    )


@step_just_wd(u'I type "([^"]*)" into the search box')
def i_type_search_term_into_the_search_box(wd, search_term):
    _send_to_search_box(wd, search_term)


@step_just_wd(u'I submit the search')
def i_submit_the_search(wd):
    _send_to_search_box(wd, '\n')


@step_just_wd(u'I expect the top result to contain the string "([^"]*)"')
def i_expect_the_top_result_to_contain_the_string_search_term(
    wd,
    search_term,
):

    @util.retry(times=5, wait=1, catch=IndexError)
    def top_result():
        return wd.find_elements_by_class_name('rc')[0]

    actual = top_result().text

    assert search_term in actual, '{!r} not in {!r}'.format(
        search_term, actual
    )



# This is done in order to apply said hooks, and may be
# changed in version 1.0 to avoid import-time behaviour
# as th8is can be confusing and results in misleading
# inspection messages such as 'Unused import statement'
#
# noinspection PyUnresolvedReferences
from fattoush import hooks



"""
Hooks that run before and after steps
"""
import functools

from lettuce import after

from fattoush.namespace import per


def if_browser(fn):
    @functools.wraps(fn)
    def _inner(step, *args, **kwargs):
        browser = per.scenario.get('browser')

        if browser is not None:
            fn(step, *args, **kwargs)

    return _inner


def if_scenario(fn):
    @functools.wraps(fn)
    def _inner(step):
        if getattr(step, 'scenario', None) is not None:
            fn(step)

    return _inner


@after.each_step
@if_browser
def hook_mark_failures(step):
    if step.failed:
        per.scenario['browser'].sauce.fail_session()



"""
Fattoush
========

Fattoush is a package that combines lettuce, webdriver and sauce to make
a tasty UI testing salad.


Running
-------

Fattoush provides its own test runner which invokes lettuce such that it
runs in saucelabs, with support of parallel test runs.

    $ fattoush --parallel=webdriver


Usage
-----

Just write your lettuce tests as normal,
but in your terrain file add

    from fattoush import hooks

You will now find that in each step you are able to add

    from fattoush import Driver

This returns a subclass of WebDriver with a class method `instance()`
which takes the current step or feature and returns the active WebDriver
instance.

License
-------
(c) 2014 Mind Candy Ltd. All Rights Reserved.
Licensed under the MIT License; you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT.
"""

from fattoush.config import VERSION
from fattoush.decorators.step import (
    screenshot as step,
    without_step as step_plain,
    with_wd as step_with_wd,
    just_wd as step_just_wd,
)
from fattoush.driver import Driver
from fattoush.namespace import (
    per,
    world,
)
from fattoush.util import retry



# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

"""
Interface between Sauce Labs and lettuce driver
"""
from abc import ABCMeta, abstractmethod
import base64
import hmac
import json
import urllib2


class SauceInterface(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def set_session_configuration(self, **conf):
        pass

    @abstractmethod
    def fail_session(self):
        pass

    @abstractmethod
    def add_file_to_storage(self, binary_string, server_path,
                            overwrite=False):
        return "sauce:storage:NAME"


class AbstractSauceBase(SauceInterface):

    __metaclass__ = ABCMeta

    session_text_template = "SauceOnDemandSessionID={0} job-name={1}"
    public_url_template = "https://saucelabs.com/jobs/{0}?auth={1}"

    user = "saucelabs_username"
    key = "saucelabs_api_key"

    default_headers = {}

    def __init__(self, config, browser):
        """
        :type config: fattoush.config.FattoushConfig
        :type browser: fattoush.driver.driver.Driver
        """

        # Bind copy of class defaults to instance
        self.default_headers = self.default_headers.copy()

        self.wd_hub = config.command_executor
        self.session_name = config.name

        self.browser = browser
        self.job_id = browser.session_id

        self.combined_key = "{0}:{1}".format(self.user, self.key)
        self.token = hmac.new(self.combined_key,
                              self.job_id).hexdigest()

        self.public_url = self.public_url_template.format(self.job_id,
                                                          self.token)

    @abstractmethod
    def request(self, endpoint, method='GET', body=None,
                extra_headers=None):
        return "Response"

    def set_session_configuration(self, **conf):
        body_content = json.dumps(conf)
        endpoint = '/rest/v1/{0}/jobs/{1}'.format(self.user,
                                                  self.job_id)

        self.request(endpoint, 'PUT', body_content)

    def fail_session(self):
        self.set_session_configuration(passed=False)

    def add_file_to_storage(self, binary_string, server_path,
                            overwrite=False):
        endpoint = '/rest/v1/storage/%s/%s?overwrite=%s' % (
            self.user.strip('/'),
            server_path.strip('/'),
            ('false', 'true')[overwrite])
        ret = self.request(endpoint=endpoint, method='POST',
                           body=binary_string)
        return "sauce-storage:%s" % server_path, ret

    @property
    def scenario_details(self):
        return {
            "user": self.user,
            "job_id": self.job_id,
            "session_name": self.session_name,
            "key": self.combined_key,
            "token": self.token,
            "public_url": self.public_url,
        }


class Local(AbstractSauceBase):
    """
    Local Interface for testing
    """
    def request(self, endpoint, method='GET', body=None,
                extra_headers=None):
        print ("Would [{0}]{1} with body of {2} (extra-headers={3})"
               .format(method, endpoint, body, extra_headers))

    def __init__(self, config, browser):
        super(Local, self).__init__(config, browser)


class Sauce(AbstractSauceBase):
    """
    Connect to real Saucelabs infrastructure
    """
    def __init__(self, config, browser):

        self.user = config.server["user"]
        self.key = config.server["key"]

        self._url_ = "http://saucelabs.com/{0}"

        super(Sauce, self).__init__(config, browser)

        self.session_text = self.session_text_template.format(
            self.job_id, self.session_name)

        self.b64_key = base64.encodestring(self.combined_key)[:-1]

        self.default_headers["Authorization"] = "Basic {0}".format(
            self.b64_key)

        print self.session_text
        print self.public_url

    def _url(self, endpoint):
        return self._url_.format(endpoint.lstrip('/'))

    def _headers(self, extra_headers):
        headers = self.default_headers.copy()
        if extra_headers is not None:
            headers.update(extra_headers)

        return headers

    def request(self, endpoint, method='GET', body=None,
                extra_headers=None):
        request = urllib2.Request(url=self._url(endpoint),
                                  data=body,
                                  headers=self._headers(extra_headers))
        request.get_method = lambda: method

        return urllib2.urlopen(request)



"""
Here we wrap usage of the lettuce world object for ease of porting
"""

from lettuce import world as _world

from fattoush.config import deprecated


def _world_attr(name, doc=None):
    return property(
        fget=lambda _: getattr(_world, name),
        fset=lambda _, value: setattr(_world, name, value),
        fdel=lambda _: delattr(_world, name),
        doc=doc
    )


class _Per(object):
    scenario = _world_attr('per_scenario')


class _World(object):
    @deprecated.FromVersion('1.0', 'Direct access prevents migration')
    def __getattr__(self, item):
        return getattr(_world, item)

    @deprecated.FromVersion('1.0', 'Direct access prevents migration')
    def __setattr__(self, key, value):
        return setattr(_world, key, value)

    @deprecated.FromVersion('1.0', 'Direct access prevents migration')
    def __delattr__(self, item):
        return delattr(_world, item)


per = _Per()
world = _World()
config = None



# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

"""
Hooks that run before and after scenarios
"""
from lettuce import (
    after,
    before,
)

from fattoush.namespace import per


@before.each_scenario
def set_per_scenario(_):
    per.scenario = {}


@before.each_scenario
def hook_rename_scenario(scenario, *_):
    feature = scenario.feature
    scenario.name = "{0}.{1}".format(feature.name, scenario.name)


def clear_browser_etc():

    browser = per.scenario.get('browser')
    if browser is not None:
        browser.delete_all_cookies()
        browser.quit()

    del per.scenario


@after.each_scenario
def clear_per_scenario(*_):
    clear_browser_etc()


@after.outline
def clear_per_iteration(*_):
    clear_browser_etc()
    set_per_scenario(_)



# -*- coding: utf-8 -*-
import json

import mock

import fattoush

from fattoush.config import FattoushConfig

from fattoush.decorators import step


@step.without_step(u'Given I pass desired capabilities of "(.*)"')
def given_i_pass_desired_capabilities_of(capabilities_json):
    fattoush.per.scenario['capabilities'] = json.loads(capabilities_json)


@step.without_step(u'When fattoush creates a configurations from this')
def when_fattoush_creates_a_configurations_from_this():
    fattoush.per.scenario['config'] = FattoushConfig(
        index=None,
        browser={
            'capabilities': fattoush.per.scenario['capabilities'],
            'options': fattoush.per.scenario.get('options', {})
        },
        server={},
        lettuce_cfg=mock.MagicMock(),
    )


@step.without_step(u'the configuration name is "([^"]*)"')
def then_the_configuration_name_is(expected):
    config = fattoush.per.scenario['config']
    actual = config.name

    assert actual == expected, 'Expected {!r} but got {!r}'.format(
        expected, actual
    )


@step.just_hashes(u'the configuration has options:')
def the_configuration_name_is(dicts):
    config = fattoush.per.scenario['config']
    options = config.options()

    expected = [dict_['option'] for dict_ in dicts]
    actual = options.arguments

    assert actual == expected, 'Expected {!r} but got {!r}'.format(
        expected, actual
    )


@step.without_step(u'Then it does not crash')
def then_it_does_not_crash():
    # This is a no-op, as it is essentially just saying that
    # the previous step did not raise an exception - if that
    # were not true we would never reach this step.
    pass


@step.just_hashes(u'And I pass browser options:')
def and_i_pass_browser_options(dicts):
    fattoush.per.scenario['options'] = [dict_['option'] for dict_ in dicts]



# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

"""
The driver class shall be a subclass of WebDriver with a classmethod
`instance()` which takes the current step or feature and returns
the active WebDriver instance.
"""

from lettuce.core import Scenario, Step
from selenium.webdriver import Remote

from fattoush import namespace
from fattoush.driver.sauce import (
    Local,
    Sauce,
)
from fattoush.namespace import per


class Driver(Remote):

    @classmethod
    def _scenario(cls, step_or_scenario):
        if isinstance(step_or_scenario, Scenario):
            return step_or_scenario.name

        if isinstance(step_or_scenario, Step):
            step = step_or_scenario

            if getattr(step, 'scenario', None) is not None:
                return step.scenario.name

            if getattr(step, 'background', None) is not None:
                return repr(step)

        raise TypeError("{0} is not an instance of {1} or {2}"
                        .format(step_or_scenario, Step, Scenario))

    @classmethod
    def instance(cls, step_or_scenario):
        """
        :rtype : Driver
        """

        if 'browser' not in per.scenario:
            per.scenario['browser'] = cls(
                namespace.config,
                cls._scenario(step_or_scenario),
            )

        return per.scenario['browser']

    @classmethod
    def got_instance(cls):
        return 'browser' in per.scenario

    @classmethod
    def kill_instance(cls):
        browser = per.scenario.pop('browser', None)

        if browser is not None:
            browser.quit()

    def __init__(self, config, name):
        """
        :type config: fattoush.config.FattoushConfig
        :type name: str
        """
        self.fattoush_config = config

        super(Driver, self).__init__(
            config.command_executor,
            config.desired_capabilities(name),
            options=config.options(),
        )

        # For saucelabs you need a user name and an auth key
        self.is_sauce = "user" in config.server

        cls = Sauce if self.is_sauce else Local
        self.sauce = cls(config=config, browser=self)



import contextlib
import functools
import warnings

import packaging.version

from fattoush.config.version import VERSION as _CURRENT


class VersionedDeprecation(Warning):
    template = "?!?"
    warning = DeprecationWarning

    def __new__(cls, version, reason):
        args = {
            'from': version,
            'current': _CURRENT,
            'reason': reason,
        }
        message = cls.template.format(**args)

        return cls.warning(message)


class NowDeprecated(VersionedDeprecation):
    template = "Deprecated since {from} (current: {current}): {reason}"
    warning = DeprecationWarning


class SoonDeprecated(VersionedDeprecation):
    template = "Will be deprecated in {from} (current: {current}): {reason}"
    warning = PendingDeprecationWarning


class FromVersion(object):
    def __init__(self, version, reason):
        self.version = packaging.version.parse(version)
        self.reason = reason

    def now(self):
        return NowDeprecated(self.version, self.reason)

    def later(self):
        return SoonDeprecated(self.version, self.reason)

    def check(self, stack_level=3):
        if _CURRENT >= self.version:
            raise self.now()

        with _always_warn():
            warnings.warn(self.later(), stacklevel=stack_level)

    def __enter__(self):
        self.check()

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __call__(self, fn):
        @functools.wraps(fn)
        def _inner(*args, **kwargs):
            self.check()
            fn(*args, **kwargs)

        return _inner


@contextlib.contextmanager
def _always_warn():
    original = warnings.filters[:]

    warnings.resetwarnings()

    yield

    warnings.filters = original



#!/usr/bin/env python
# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

"""
This script uses fattoush to run lettuce
"""


def console():
    from fattoush.config import FattoushConfigGroup

    config = FattoushConfigGroup.from_cli_args()
    failures = config.run()

    if failures:
        raise SystemExit(failures)


if __name__ == '__main__':
    console()



# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

"""
`parse_args` is the One True Place to see how the arguments passed in
at the command line translate to what is passed into the
FattoushConfigGroup, which in turn controls how everything else is run.
"""
import os
import argparse


# noinspection SpellCheckingInspection
def add_lettuce_group(parser):
    """
    :type parser: argparse.ArgumentParser
    """
    group = parser.add_argument_group('lettuce arguments')
    group.add_argument("-v", "--verbosity",
                       dest="verbosity",
                       default=4,
                       help='The verbosity level')

    group.add_argument("-s", "--scenarios",
                       dest="scenarios",
                       default=None,
                       help='Comma separated list of scenarios to run')

    group.add_argument("-t", "--tag",
                       dest="tags",
                       action='append',
                       default=[],
                       help='Tells lettuce to run the specified tags '
                            'only; can be used multiple times to '
                            'define more tags(prefixing tags with "-" '
                            'will exclude them and prefixing with "~" '
                            'will match approximate words)')

    group.add_argument("-r", "--random",
                       dest="random",
                       action="store_true",
                       default=False,
                       help="Run scenarios in a more random order to "
                            "avoid interference")

    group.add_argument("--with-xunit",
                       dest="enable_xunit",
                       action="store_true",
                       default=False,
                       help='Output JUnit XML test results to a file')

    group.add_argument("--xunit-file",
                       dest="xunit_file",
                       default=None,
                       type=str,
                       help='Write JUnit XML to this file. Defaults to'
                            ' lettucetests.xml')

    group.add_argument("--failfast",
                       dest="failfast",
                       default=False,
                       action="store_true",
                       help='Stop running in the first failure')

    group.add_argument("--pdb",
                       dest="auto_pdb",
                       default=False,
                       action="store_true",
                       help='Launches an interactive debugger upon '
                            'error')

    # Inspection assumes os.curdir is the project directory, in
    # practice it should be the working directory when fattoush is run.
    # noinspection PyUnresolvedReferences
    group.add_argument('base_path',
                       default=os.path.join(os.curdir, 'features'),
                       nargs='?')


def add_fattoush_group(parser):
    """
    :type parser: argparse.ArgumentParser
    """
    group = parser.add_argument_group('fattoush',
                                      'Arguments that affect how '
                                      'fattoush is run')
    group.add_argument('--parallel',
                       default=None,
                       choices=['webdriver'],
                       help='Use this to run multiple sets of tests '
                            'in parallel. Can specify to fork on '
                            '{%(choices)s}',
                       metavar='SPLIT_ON')

    group.add_argument('-c', '--config-file',
                       dest='config', action='store',
                       default=None, metavar='JSON_FILE_PATH',
                       help='Unless this is specified, the desired '
                            'webdriver configuration will be read '
                            'from the environment, for compatibility '
                            'with Teamcity\'s Sauce Labs Build '
                            'Feature functionality. '

                            'Given the path to a file fattoush will '
                            'read a json configuration.')

    printing = parser.add_mutually_exclusive_group()

    printing.add_argument('--ps', '--print-schema',
                          dest='print_schema', action='store_true',
                          help='Print to stdout the schema which '
                               'config files must adhere to.')

    printing.add_argument('--pc', '--print-example-config',
                          dest='print_config', action='store_true',
                          help='Print to stdout an example config file')


def parse_args(args):
    from ..config import VERSION
    parser = argparse.ArgumentParser(prog='fattoush')
    parser.add_argument('--version',
                        action='version',
                        version=VERSION)

    add_fattoush_group(parser)
    add_lettuce_group(parser)

    return parser.parse_args(args)



import datetime
import functools
import os.path
import time

import future.utils

LOGFILE_NAME_TEMPLATE = (
    "{datetime:%Y%m%d_%H%M%S.%f}{parent_name} - {sentence}.{ext}"
)


def truncate(long_string, max_length, link_with='...'):
    if len(long_string) <= max_length:
        return long_string

    text_length = max_length - len(link_with)
    end_size = int(text_length / 4)
    start_size = text_length - end_size

    return "{start}{link}{end}".format(
        start=long_string[:start_size],
        end=long_string[-end_size:],
        link=link_with,
    )


def filename_for_step(
    step,
    ext,
    template=LOGFILE_NAME_TEMPLATE,
    max_characters=200,
    **extra_format_string_kwargs
):
    """
    On some platforms filename length is limited to 200 characters. In
    order to work with this let's trim down a few parts to try and keep
    the file names useful.
    """
    parent = step.parent
    parent_name = getattr(parent, 'name', None)
    now = datetime.datetime.now()

    mandatory_characters = len(template.format(
        datetime=now,
        parent_name='',
        sentence='',
        ext=ext,
        **extra_format_string_kwargs
    ))

    if parent_name is None:  # Must be a background
        parent_name = parent.feature.name

    characters_left = max_characters - mandatory_characters

    parent_name = truncate(parent_name, characters_left / 2)

    characters_left -= len(parent_name)

    sentence = truncate(step.sentence, characters_left)

    return template.format(
        sentence=sentence,
        parent_name=parent_name,
        datetime=now,
        ext=ext,
        **extra_format_string_kwargs
    )


def filename_in_created_dir(dir_name, step, ext, **kwargs):
    dir_name = os.path.abspath(dir_name)

    if not os.path.exists(dir_name):
        os.makedirs(dir_name)

    return os.path.join(dir_name, filename_for_step(step, ext=ext, **kwargs))


def try_map(fn, args_iterable):
    ex = None
    failures = []

    for args in args_iterable:
        try:
            yield fn(args)
        except Exception as ex:
            print(ex)

            failures.append((ex, args))

    if ex is not None:
        msg = '\n\t'.join(
            ['Failed to map {} against all args:'.format(fn)] +
            [
                '{}({}) failed with {}'.format(fn, args, ex)
                for ex, args in failures
            ]
        )

        _runtime_from(msg, ex)


def _runtime_from(msg, cause):
    return future.utils.raise_from(RuntimeError(msg), cause)


def retry(times=1, wait=0, catch=RuntimeError):

    def decorator(fn):
        @functools.wraps(fn)
        def _inner(*args, **kwargs):
            ex = None

            for _ in xrange(times):
                try:
                    return fn(*args, **kwargs)
                except catch as ex:
                    time.sleep(wait)

            if ex is not None:
                msg = '{}(*{}. **{}) failed {} times'.format(
                    fn, args, kwargs, times
                )
                _runtime_from(msg, ex)

        return _inner

    return decorator


def _attrs(obj, include_private=False):
    """ Iterate over a namespace as name-value pairs """
    for name in dir(obj):
        if name.startswith('_') and not include_private:
            continue

        yield name, getattr(obj, name)

def flatmap(obj, to_cls=None):
    """
    Flatmap a decorated function over a namespace

    In the absence of `to_cls` this creates a generator. With
    `to_cls` that generator will be exhausted into an instance
    of that class.

    This is useful for creating say a dictionary within a
    class' namespace without polluting it with extra mess
    """
    def iter_(fn):
        for name, value in _attrs(obj):
            for item in fn(name, value):
                yield item

    if to_cls is None:
        return iter_

    def do(fn):
        return to_cls(iter_(fn))

    return do



"""
Functions and classes for taking commandline arguments and forming an
object which will give you a set of FattoushConfig objects, each of
which contains all the information necessary to run lettuce against a
specified webdriver configuration, whither in saucelabs or local.
"""
import copy
import json
import multiprocessing
import urlparse
from os import environ, path

from jsonschema import validate

from fattoush import namespace, util
from fattoush.config import config, deprecated
from fattoush.runner.parsing import parse_args


class FattoushConfigGroup(object):

    schema = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "description": "schema for a single file fattoush config",

        "type": "object",
        "required": ["browsers"],
        "properties": {
            "description": {"$ref": "#/definitions/comment"},
            "server": {
                "anyOf": [
                    {"$ref": "#/definitions/saucelabs"},
                    {"$ref": "#/definitions/local"}
                ]
            },
            "browsers": {
                "anyOf": [
                    {"$ref": "#/definitions/browsers"},
                    {"$ref": "#/definitions/browsers-pre-0.4"}
                ]
            }
        },
        "definitions": {
            "comment": {
                "description":
                    "Not used anywhere, just a comment, has no validation, "
                    "so you can store anything, intended to allow strings, "
                    "and lists of strings, in order to facilitate multi-line "
                    "comments, but could take an object with all sorts "
                    "of meta-data... Please don't abuse."
            },
            "saucelabs": {
                "description": "Specification of how to connect to saucelabs",
                "properties": {
                    "description": {"$ref": "#/definitions/comment"},
                    "url": {
                        "description":
                            "The initial URL to load when the test begins",
                        "type": "string"
                    },
                    "user": {
                        "description":
                            "The user name used to invoke Sauce OnDemand",
                        "type": "string"
                    },
                    "key": {
                        "description":
                            "The access key for the user used to invoke "
                            "Sauce OnDemand",
                        "type": "string"
                    }
                },
                "required": ["user", "key"],
                "additionalProperties": False
            },
            "local": {
                "description":
                    "Specification of how to connect to a selenium server on "
                    "your local network. Defaults to 127.0.0.1:4444",
                "properties": {
                    "description": {"$ref": "#/definitions/comment"},
                    "host": {
                        "description": "The hostname of the Selenium server",
                        "type": "string"
                    },
                    "port": {
                        "description": "The port of the Selenium server",
                        "type": "string"
                    },
                    "url": {
                        "description":
                            "The initial URL to load when the test begins",
                        "type": "string"
                    }
                },
                "additionalProperties": False
            },
            "browsers-pre-0.4": {
                "type": "array",
                "minItems": 1,
                "items": {"$ref": "#/definitions/capabilities"},
                "uniqueItems": True,
            },
            "browsers": {
                "description":
                    "Specification of the browsers to ask webdriver to open",
                "type": "object",
                "properties": {
                    "capabilities": {
                        "description":
                            "Keyed lookup of desired capabilities",
                        "type": "object",
                        "additionalProperties": {
                            "$ref": "#/definitions/capabilities",
                        },
                    },
                    "options": {
                        "description":
                            "Keyed lookup of browser options",
                        "type": "object",
                        "additionalProperties": {
                            "$ref": "#/definitions/options",
                        },
                    },
                    "selection": {
                        "type": "array",
                        "items": {"$ref": "#/definitions/browser"},
                    },
                },
                "required": ["selection"],
            },
            "capabilities": {
                "description":
                    "Specification of the desired capabilities to "
                    "give webdriver",
                "type": "object",
                "properties": {
                    "description": {"$ref": "#/definitions/comment"},
                    "platform": {"type": "string"},
                    "os": {"type": "string"},
                    "browser": {
                        "type": "string",
                        "enum": list(config.FattoushConfig.desired.keys()),
                    },
                    "url": {
                        "type": "string",
                        "description":
                            "Contains the operating system, version and "
                            "browser name of the selected browser, in a "
                            "format designed for use by the Selenium "
                            "Client Factory"
                    },
                    "browser-version": {"type": "string"}
                },
                "required": ["browser"]
            },
            "options": {
                "description":
                    "Specification of the browser options to give webdriver",
                "type": "array",
                "items": {"type": "string"},
            },
            "browser": {
                "description":
                    "Specification of a browsers to ask webdriver to open",
                "type": "object",
                "properties": {
                    "description": {"$ref": "#/definitions/comment"},
                    "capabilities": {
                        "anyOf": [
                            {"$ref": "#/definitions/capabilities"},
                            {
                                "type": "string",
                                "description": "Key within capabilities",
                            },
                        ]
                    },
                    "options": {
                        "anyOf": [
                            {"$ref": "#/definitions/options"},
                            {
                                "type": "string",
                                "description": "Key within options",
                            },
                        ]
                    },
                },
            },
        }
    }

    @staticmethod
    def config_from_env():
        """
        Joyfully SauceConnect presents completely different
        environmental variables based on whether you are
        running against one saucelabs session or several.

        This function will return a list of different session
        configurations whatever the case - empty if there are
        none, a list with only one item if there is only one,
        or a list of multiple items if there are many. It's
        almost as if this would have been the sensible way
        for them to do it too.

        The only part which is different in the singular case is
        the missing 'os' key.

        in the multiple case the json string is documented to be
        of the format as follows:

        [
            {
                "platform":"LINUX",
                "os":"Linux",
                "browser":"firefox",
                "url":"sauce-ondemand:?os=Linux&
                                       browser=firefox&
                                       browser-version=16",
                "browser-version":"16"
            },
            {
                "platform":"VISTA",
                "os":"Windows 2008",
                "browser":"iexploreproxy",
                "url":"sauce-ondemand:?os=Windows 2008&
                                       browser=iexploreproxy&
                                       browser-version=9",
                "browser-version":"9"
            }
        ]
        """

        try:
            json_data = environ.get("SAUCE_ONDEMAND_BROWSERS")
            browsers = json.loads(json_data)
        except (ValueError, TypeError):
            url = environ.get("SELENIUM_DRIVER")
            try:
                query = urlparse.urlparse(url).query
                parsed = urlparse.parse_qs(query)
            except AttributeError:
                parsed = {}
            browsers = [
                {
                    "platform": environ.get("SELENIUM_PLATFORM"),
                    "browser": environ.get("SELENIUM_BROWSER"),
                    "url": url,
                    "browser-version": environ.get("SELENIUM_VERSION")
                }
            ]
            if "os" in parsed:
                browsers[0]["os"] = parsed["os"]

        return {
            "server": {
                "host": environ.get("SELENIUM_HOST"),
                "port": environ.get("SELENIUM_PORT"),
                "url": environ.get("SELENIUM_URL"),
                "user": environ.get("SAUCE_USER_NAME"),
                "key": environ.get("SAUCE_API_KEY")
            },
            "browsers": browsers
        }

    @staticmethod
    def config_from_file(absolute_file_path):
        """ Supports reading config a single json file. """
        return json.load(open(absolute_file_path))

    @property
    def xrange(self):
        return xrange(len(self.configs["browsers"]))

    @classmethod
    def from_cli_args(cls):
        import sys
        options = parse_args(sys.argv[1:])
        return cls(options)

    def __init__(self, options):
        """
        Takes the options that are passed into the runner and
        creates a config object that can be referred to throughout
        fattoush.

        :type options: Namespace
        """
        if options.print_schema:
            print(json.dumps(self.schema, indent=2, sort_keys=True))
            exit(0)
        elif options.print_config:
            file_name = path.join(path.dirname(__file__),
                                  'example_config.json')
            with open(file_name) as example:
                print example.read()
            exit(0)

        self._raw_options = options
        self.parallel = options.parallel

        if options.config is None:
            self.configs = self.config_from_env()
        else:
            self.configs = self.config_from_file(options.config)

        validate(self.configs, self.schema)

        self._convert_legacy_config()

        xunit_filename = ('lettucetests.xml'
                          if options.enable_xunit
                          and options.xunit_file is None
                          else options.xunit_file)

        self.lettuce_options = {
            'base_path': options.base_path,
            'scenarios': options.scenarios,
            'verbosity': options.verbosity,
            'random': options.random,
            'enable_xunit': options.enable_xunit,
            'xunit_filename': xunit_filename,
            'failfast': options.failfast,
            'auto_pdb': options.auto_pdb,
            'tags': ([tag.strip('@') for tag in options.tags]
                     if options.tags else None)
        }

    def _convert_legacy_config(self):
        if isinstance(self.configs['browsers'], list):
            # Before version 0.4 all capabilities went
            # straight in the browser objects, since 0.4 this
            # has become nested to also include options to set
            # on the remote executor. The pre-0.4 format shall
            # be deprecated from 0.5 onwards.
            with deprecated.FromVersion('0.5', 'Legacy browser config format'):

                browsers = self.configs["browsers"]

                self.configs["browsers"] = {
                    'selection': [
                        {'capabilities': browser} for browser in browsers
                    ]
                }

    @property
    def to_dict(self):
        """
        The returned dictionary gives a shallow copy of the data
        required to create a FeatureConfig.
        """
        return {
            'lettuce_options': self.lettuce_options.copy(),
            'config': self.configs.copy()
        }

    def run(self):
        if not self.configs:
            raise IndexError('There are no webdriver configs against '
                             'which to run lettuce.')

        runner = (
            self._run_parallel if self.parallel == 'webdriver' else
            self._run_series
        )

        return sum(
            result.scenarios_ran - result.scenarios_passed
            for result in runner()
            if result.features_passed < result.features_ran
        )

    def _run_series(self):
        """
        Runs lettuce against each browser configuration one at a time

        :type self: fattoush.config.FattoushConfigGroup
        """
        return list(
            util.try_map(
                _run_kwargs,
                self._iter_browser_kwargs(),
            )
        )

    def _run_parallel(self):
        """
        Runs lettuce against all the browser configurations at the same
        time in different processes.

        :type self: fattoush.config.FattoushConfigGroup
        """
        multiprocessing.Pool().map(
            _run_kwargs,
            self._iter_browser_kwargs(),
        )

    def _iter_browser_kwargs(self):
        browsers = self.configs["browsers"]

        kwargs = {
            'server': self.configs.get("server", {}),
            'lettuce': self.lettuce_options,
        }

        for (index, browser) in enumerate(browsers['selection']):

            kwargs['index'] = index

            caps = browser.setdefault('capabilities', {})
            if not isinstance(caps, dict):
                browser['capabilities'] = browsers['capabilities'][caps]

            opts = browser.setdefault('options', [])
            if not isinstance(opts, list):
                browser['options'] = browsers['options'][opts]

            kwargs['browser'] = browser

            yield kwargs


def _run_kwargs(kwargs):
    """ This would be a starmap in py3... """
    return run_single(**copy.deepcopy(kwargs))


def run_single(index, browser, server, lettuce):
    """
    :type index: int
    :type browser: dict
    :type server: dict
    :type lettuce: dict
    """

    fattoush_config = config.FattoushConfig(
        index=index,
        browser=browser,
        server=server,
        lettuce_cfg=lettuce,
    )
    namespace.config = fattoush_config

    try:
        result = fattoush_config.run()
        return result
    finally:
        namespace.config = None



"""
Functions for putting together sauce configuration,
be it from environmental variables or from a JSON string.
"""
import json

from os import path

import lettuce
import selenium.webdriver

import fattoush.util


class FattoushConfig(object):
    _capabilities = None

    @fattoush.util.flatmap(selenium.webdriver.DesiredCapabilities, to_cls=dict)
    def desired(cap_name, cap):
        """
        :type cap_name:    str
        :type cap:         str
        """
        if isinstance(cap, dict):
            browser_name = cap.get("browserName")

            if browser_name is not None:

                cap_name_lower = cap_name.lower()

                yield cap_name_lower, cap

                if (
                    cap_name_lower != browser_name and
                    cap_name_lower == browser_name.lower
                ):
                    yield browser_name, cap

    desired.update({
        "googlechrome": selenium.webdriver.DesiredCapabilities.CHROME,
        "phantom": selenium.webdriver.DesiredCapabilities.PHANTOMJS,
        "iexploreproxy": selenium.webdriver.DesiredCapabilities.INTERNETEXPLORER,
    })

    @fattoush.util.flatmap(selenium.webdriver, to_cls=dict)
    def _options(_, cls):
        try:
            if cls.__name__ == 'Options':
                option = cls()

                yield option.capabilities["browserName"], cls
        except (NameError, KeyError):
            pass

    def _augment_xunit_filename(self):
        index = self.index
        xunit_file = self.run_args['xunit_filename']

        if xunit_file is None:
            return

        if '.' in path.basename(xunit_file):
            prefix, suffix = path.splitext(xunit_file)
            xunit_file = '{0}_{1}.{2}'.format(prefix, index, suffix)
        else:
            xunit_file = '{0}_{1}'.format(xunit_file, index)
        self.run_args['xunit_filename'] = xunit_file

    def _augment_server(self):
        if self.server is None:
            self.server = {}
        server = self.server
        if "host" not in server:
            server[u"host"] = ("127.0.0.1" if "user" not in server
                               else "ondemand.saucelabs.com")
        if "port" not in server:
            server[u"port"] = ("4444" if "user" not in server
                               else "80")
        if "url" in server and not server["url"]:
            del server["url"]

    def run(self):
        try:
            return lettuce.Runner(**self.run_args).run()
        except BaseException:
            print "Lettuce raised the following exception:"
            raise

    def _augment_browser(self):
        """
        All parts of the browser config must be a string except for the
        description, which can hold any format.
        """

        def to_string(obj):
            try:
                if isinstance(obj, unicode):
                    return obj
                elif isinstance(obj, list):
                    return ''.join(obj)
            except TypeError:
                return json.dumps(obj)

        if "description" in self.browser:
            self.browser["description"] = to_string(
                self.browser["description"])

    def __init__(self, index, browser, server, lettuce_cfg):
        """
        :type index: int
        :type browser: dict
        :type server: dict
        :type lettuce_cfg: dict
        """
        self.index = index
        self.server = server
        self.browser = browser.copy()
        self.run_args = lettuce_cfg

        self._augment_xunit_filename()
        self._augment_server()

        self._augment_browser()
        self.name = _create_name_from_capabilities(
            self.browser['capabilities'],
            surround=False,
        )

    @property
    def command_executor(self):
        host = self.server["host"].rstrip('/')

        try:
            protocol, host = host.split("://")
        except ValueError:
            protocol = 'http'

        try:
            host, url_path = host.split(":")
        except ValueError:
            port = self.server["port"]
        else:
            split_path = url_path.split("/")
            port = split_path[0]
            split_path[0] = host
            host = '/'.join(split_path)

        split_path = host.split('/')
        host, split_path = split_path[0], split_path[1:]

        endpoint = "/".join(split_path) if split_path else "wd/hub"

        if "user" in self.server and "key" in self.server:
            return ('{protocol}://{user}:{key}@{host}:{port}/{endpoint}'
                    ).format(protocol=protocol,
                             user=self.server["user"],
                             key=self.server["key"],
                             host=host,
                             port=port,
                             endpoint=endpoint)

        elif "user" in self.server:
            return ('{protocol}://{user}@{host}:{port}/{endpoint}'
                    ).format(protocol=protocol,
                             user=self.server["user"],
                             host=host,
                             port=port,
                             endpoint=endpoint)
        else:
            return ('{protocol}://{host}:{port}/{endpoint}'
                    ).format(protocol=protocol,
                             host=host,
                             port=port,
                             endpoint=endpoint)

    @property
    def base_capabilities(self):
        browser_name = self.browser['capabilities']["browser"]

        return self.desired.get(
            browser_name,
            selenium.webdriver.DesiredCapabilities.CHROME
        )

    def desired_capabilities(self, name):
        capabilities = self.base_capabilities.copy()

        capabilities.update({
            k: v for k, v in self.browser['capabilities'].items()
            if k != "browser"
        })

        capabilities["name"] = name

        return capabilities

    def options(self):
        browser_name = self.base_capabilities["browserName"]

        cls = self._options.get(browser_name)

        if cls is not None:

            browser_options = cls()

            for option in self.browser['options']:
                browser_options.add_argument(option)

            return browser_options


def _create_name_from_capabilities(caps, surround=True):
    if isinstance(caps, (str, unicode)):
        return caps

    if hasattr(caps, 'iteritems'):
        return _do_join(
            parts=_mapping_to_name_parts(caps),
            surround=surround
        )
    elif hasattr(caps, '__iter__'):
        return _do_join(
            parts=_iterable_to_name_parts(caps),
            surround=surround
        )

    return repr(caps)


def _mapping_to_name_parts(caps):
    for key, value in sorted(caps.items()):
        if key == 'description':
            continue

        yield '='.join((key, _create_name_from_capabilities(value)))


def _iterable_to_name_parts(caps):
    for value in caps:
        yield _create_name_from_capabilities(value)


def _do_join(parts, surround, with_=';'):
    whole = with_.join(parts)

    if surround:
        whole = '{{{}}}'.format(whole)

    return whole



# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

from setuptools import setup, find_packages
from os import path
import sys


sys.path.insert(0, path.join(path.dirname(__file__), 'src'))
try:
    import fattoush
except ImportError:
    class fattoush:
        """ Placeholder for tox """

        VERSION = "Placeholder for tox"


setup(
    name='Fattoush',
    package_data={'': ['*.txt']},
    author='Alistair Broomhead',
    version=str(fattoush.VERSION),
    author_email='alistair.broomhead+python@gmail.com',
    description="A delicious testing framework",
    license='MIT',
    url='https://github.com/alistair-broomhead/fattoush',
    long_description=fattoush.__doc__,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=[
        "argparse>=1.2.1",
        "future",
        "jsonschema>=2.3.0",
        "lettuce==0.2.19",
        "packaging",
        "selenium>=3.0.0",
    ],
    extras_require={
        "fast": [
            "python-Levenshtein",
        ]
    },
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'fattoush = fattoush.runner.bin:console']
    }
)



# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

"""
Hooks that run before and after features
"""

from lettuce import before

from fattoush import namespace


@before.each_feature
def hook_before_feature(feature):
    """
    Change the feature name to include information about where it is
    running - this will appear in the Saucelabs session and so
    having this information makes it easier to see where any
    unexpected behaviour may have occurred.
    """
    config = namespace.config
    feature.name = "[{0}] {1}".format(config.name, feature.name)
